# flake8: noqa
import os
import tempfile
import warnings
from glob import glob
from pathlib import Path

import numpy as np
import pandas as pd
from astropy.utils.data import CacheMissingWarning, cache_contents
from astropy.utils.data import clear_download_cache as _astropy_clear_download_cache
from astropy.utils.data import download_file, import_file_to_cache, is_url_in_cache
from tqdm import tqdm

from . import log

KERNELS = {
    "naif0012.tls": "https://naif.jpl.nasa.gov/pub/naif/generic_kernels/lsk/",
    "de440.bsp": "https://naif.jpl.nasa.gov/pub/naif/generic_kernels/spk/planets/",
    "kplr2012299191755.tsc": "https://archive.stsci.edu/missions/kepler/spice/",
    "kplr2013205175543.tsc": "https://archive.stsci.edu/missions/kepler/spice/",
    "spk_2013161000000_2013163203537_kplr.bsp": "https://archive.stsci.edu/missions/kepler/spice/",
    "spk_2013358000000_2013360154555_kplr.bsp": "https://archive.stsci.edu/missions/kepler/spice/",
    "kplr2018134232543.tsc": "https://archive.stsci.edu/missions/k2/spice/",
    "spk_2018290000000_2018306220633_kplr.bsp": "https://archive.stsci.edu/missions/k2/spice/",
    "tess2018236164754-41096_sclk.tsc": "https://archive.stsci.edu/missions/tess/engineering/",
    "tess_20_year_long_predictive.bsp": "https://archive.stsci.edu/missions/tess/models/",
}


def get_tess_bsp():
    """Get all the listings of the spice kernels for TESS, including any new files."""
    df = pd.read_csv("https://archive.stsci.edu/missions/tess/models/", header=None)
    tess_bsp = {
        l.split(">")[2].split("<")[0]: "https://archive.stsci.edu/missions/tess/models/"
        for l in df[0]
        if "TESS_EPH_DEF" in l
    }
    return tess_bsp


KERNELS.update(get_tess_bsp())


def get_file_path(url):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", CacheMissingWarning)
        return download_file(
            url, cache=True, show_progress=False, pkgname="lkspacecraft"
        )


def get_file_paths():
    """Ensure the file is downloaded and valid using Astropy's download_file."""
    file_paths = []
    file_names = list(KERNELS.keys())
    progress_bar = None
    for idx, file_name in enumerate(file_names):
        log.debug(f"Finding {file_name}.")
        url = KERNELS[file_name]
        if is_url_in_cache(url + file_name, pkgname="lkspacecraft"):
            file_paths.append(get_file_path(url + file_name))
            log.debug(f"Found {file_name} in cache.")
            continue
        if progress_bar is None:
            progress_bar = tqdm(
                range(0, len(file_names)),
                initial=idx,
                total=len(file_names),
                desc="Downloading SPICE Kernels",
            )
        file_paths.append(get_file_path(url + file_name))
        progress_bar.n = idx
        progress_bar.refresh()
        log.debug(f"Downloaded {file_name}.")
    return file_paths


def clear_download_cache():
    _astropy_clear_download_cache(pkgname="lkspacecraft")


def truncate_directory_string(directory_string):
    """Turns a directory string into a SPICE compliant list of directorys..."""
    lines = []
    line = ""
    for word in directory_string.split("/"):
        if word == "":
            continue
        if len(line) < 130:
            line = f"{line}/{word}"
        else:
            line = f"{line}+"
            lines.append(line)
            line = ""
    lines.append(line)
    return lines


META_START = """KPL/MK

lkspacecraft meta kernel
==============

The generic kernels listed below can be obtained from NAIF generic kernels:
    https://naif.jpl.nasa.gov/pub/naif/generic_kernels/
The Kepler kernels below can be obtained from MAST
    https://archive.stsci.edu/missions/kepler/spice/
The K2 kernels below can be obtained from MAST
    https://archive.stsci.edu/missions/k2/spice/
The TESS kernels below can be obtained from MAST
    https://archive.stsci.edu/missions/tess/engineering/
    https://archive.stsci.edu/missions/tess/models/

\\begindata

"""
META_END = """

\\begintext   
"""


def create_meta_kernel():
    """Create a meta kernel out of the cached SPICE kernels"""
    paths = get_file_paths()
    if len(paths) == 0:
        raise ValueError(
            "Can not find any SPICE kernels. Check documentation on installation."
        )
    cache_dirs = np.unique([os.path.dirname(os.path.dirname(f)) for f in paths])
    if len(cache_dirs) != 1:
        raise ValueError(
            "You have provided multiple cache directories for SPICE kernels, try reinstalling."
        )

    path_values = truncate_directory_string(cache_dirs[0])
    path_symbols = ["cache"]
    kernels_to_load = ["$cache/" + path[len(cache_dirs[0]) + 1 :] for path in paths]

    def format_list(l, pad=10):
        if len(l) == 0:
            return ""
        if len(l) == 1:
            return f" '{l[0]}'"
        output = f" '{l[0]}'"
        for i in l[1:]:
            output += "\n" + "".join([" "] * pad) + "'" + i + "'"
        return output

    output = f"""{META_START}
    \n    PATH_VALUES = ({format_list(path_values, 20)}              )
    \n    PATH_SYMBOLS = ({format_list(path_symbols, 21)}              )
    \n    KERNELS_TO_LOAD = ({format_list(kernels_to_load, 24)}              )
    {META_END}
    """

    # Create a temporary file
    with tempfile.NamedTemporaryFile(mode="w+", delete=False) as temp_file:
        temp_file.write(output)
        # Get the file name
        temp_file_name = temp_file.name

    import_file_to_cache(
        "https://github.com/lightkurve/lkspacecraft/src/lkspacecraft/data/Meta.txt",
        temp_file_name,
        pkgname="lkspacecraft",
    )
    return


# def create_meta_kernel():
#     """Function that makes a meta kernel text file in a directory with a reasonable order.

#     We assume that everything in KERNELDIR/generic is required, and has higher priority than the mission kernels.
#     """

#     META_START = """KPL/MK

# lkspacecraft meta kernel
# ==============

#     The generic kernels listed below can be obtained from NAIF generic kernels:
#         https://naif.jpl.nasa.gov/pub/naif/generic_kernels/
#     The Kepler kernels below can be obtained from MAST
#         https://archive.stsci.edu/missions/kepler/spice/
#     The K2 kernels below can be obtained from MAST
#         https://archive.stsci.edu/missions/k2/spice/
#     The TESS kernels below can be obtained from MAST
#         https://archive.stsci.edu/missions/tess/engineering/
#         https://archive.stsci.edu/missions/tess/models/

#     \\begindata

#     """
#     META_END = """

#     \\begintext
#     """
#     path_values = []
#     path_symbols = []
#     kernels_to_load = []
#     for dirname in glob(f"{KERNELDIR}*"):
#         for d in truncate_directory_string(dirname):
#             path_values.append(d)
#         path_symbols.append(dirname.split("/")[-1])
#         for d in np.sort(glob(dirname + "/*")):
#             kernels_to_load.append("$" + dirname.split("/")[-1] + d[len(dirname) :])

#     def format_list(l, pad=10):
#         if len(l) == 0:
#             return ""
#         if len(l) == 1:
#             return f" '{l[0]}'"
#         output = f" '{l[0]}'"
#         for i in l[1:]:
#             output += "\n" + "".join([" "] * pad) + "'" + i + "'"
#         return output

#     output = f"""{META_START}
#     \n    PATH_VALUES = ({format_list(path_values, 20)}              )
#     \n    PATH_SYMBOLS = ({format_list(path_symbols, 21)}              )
#     \n    KERNELS_TO_LOAD = ({format_list(kernels_to_load, 24)}              )
#     {META_END}
#     """
#     with open(f"{PACKAGEDIR}/data/Meta.txt", "w") as file:
#         file.write(output)
#     return



# Standard library
import os  # noqa

PACKAGEDIR = os.path.abspath(os.path.dirname(__file__))

from importlib.metadata import PackageNotFoundError, version  # noqa


def get_version():
    try:
        return version("lkspacecraft")
    except PackageNotFoundError:
        return "unknown"


__version__ = get_version()

import logging  # noqa: E402
import os  # noqa
from glob import glob  # noqa

log = logging.getLogger("lkspacecraft")

PACKAGEDIR = os.path.abspath(os.path.dirname(__file__))
KERNELDIR = f"{PACKAGEDIR}/data/kernels/"

# from .io import update_kernels
from .utils import create_meta_kernel  # noqa

# update_kernels() # This function should grab kernels and update the kernel directory
create_meta_kernel()

from .spacecraft import KeplerSpacecraft, TESSSpacecraft  # noqa



"""Classes for working with the orbits of spacecraft"""

from typing import Union

import astropy.units as u
import numpy as np
import numpy.typing as npt
import spiceypy
from astropy.constants import c
from astropy.time import Time
from astropy.utils.data import cache_contents


class BadEphemeris(Exception):
    def __init__(self, message):
        super().__init__(message)


class Spacecraft(object):
    """
    A base class for handling spacecraft ephemeris data and calculations.

    This class provides methods to retrieve spacecraft position, velocity, light travel time,
    and related calculations using SPICE kernels. It supports transformations between
    time formats and computes barycentric time corrections and velocity aberration effects.

    Attributes
    ----------
    start_time : astropy.time.Time
        The start time of the loaded SPICE kernel data.
    end_time : astropy.time.Time
        The end time of the loaded SPICE kernel data.

    Methods
    -------
    get_spacecraft_position(time, observer="SOLAR SYSTEM BARYCENTER")
        Returns the position vector (x, y, z) in kilometers relative to the specified observer.
    get_spacecraft_velocity(time, observer="SOLAR SYSTEM BARYCENTER")
        Returns the velocity vector (vx, vy, vz) in kilometers per second relative to the observer.
    get_spacecraft_light_travel_time(time, observer="SOLAR SYSTEM BARYCENTER")
        Computes the one-way light travel time to the observer in seconds.
    get_barycentric_time_correction(time, ra, dec)
        Calculates the barycentric time correction for a target specified by RA and Dec.
    get_velocity_aberrated_positions(time, ra, dec)
        Computes the RA and Dec of a target after applying velocity aberration.
    """

    def __init__(self):
        """
        Initializes the Spacecraft object and loads SPICE kernels.

        This method clears any previously loaded SPICE kernels, loads the kernels specified
        in the `Meta.txt` file, and determines the start and end times of the kernel data.

        Raises
        ------
        Exception
            If there is an issue loading the SPICE kernels or retrieving the kernel time coverage.
        """
        meta_kernel = cache_contents(pkgname="lkspacecraft")[
            "https://github.com/lightkurve/lkspacecraft/src/lkspacecraft/data/Meta.txt"
        ]
        spiceypy.kclear()
        spiceypy.furnsh(meta_kernel)
        self.start_time, self.end_time = self._get_kernel_start_and_end_times()

    def _get_kernel_start_and_end_times(self):
        # Get a list of loaded kernels
        kernel_list = spiceypy.ktotal("ALL")  # Get the count of all loaded kernels
        start_et = float("inf")
        end_et = float("-inf")

        for i in range(kernel_list):
            kernel_name = spiceypy.kdata(i, "ALL")[0]

            # Check if the kernel is SPK or CK to calculate coverage
            kernel_type = spiceypy.kdata(i, "ALL")[1]

            if kernel_type in ["SPK", "CK"]:
                # Create a window for coverage
                coverage_window = spiceypy.stypes.SPICEDOUBLE_CELL(2**10)

                # Query coverage for the specific kernel
                try:
                    if kernel_type == "SPK":
                        spiceypy.spkcov(
                            kernel_name, self.spacecraft_code, coverage_window
                        )  # Replace with your NAIF ID
                    else:
                        continue

                    # Extract start and end times for the current kernel
                    interval_start = spiceypy.wnfetd(coverage_window, 0)[0]
                    interval_end = spiceypy.wnfetd(coverage_window, 0)[1]

                    # Update the global start and end times
                    start_et = min(start_et, interval_start)
                    end_et = max(end_et, interval_end)
                except Exception:
                    continue
        start_time = Time(spiceypy.et2datetime(start_et))
        end_time = Time(spiceypy.et2datetime(end_et))
        return start_time, end_time

    def __repr__(self):
        return "KeplerSpacecraft"

    def _process_time(self, time: Time):
        # convert to astropy.time.Time if needed
        if not isinstance(time, Time):
            try:
                time = Time(time, format="jd")
            except ValueError:
                try:
                    time = Time(time)
                except Exception:
                    raise ValueError(
                        "Can not parse input time. Pass an `astropy.time.Time` object."
                    )
        if time.ndim == 0:
            time = Time([time])
        return time

    def _get_state_vector(self, time: Time, observer="SOLAR SYSTEM BARYCENTER"):
        time = self._process_time(time)
        et = spiceypy.str2et(time.isot)
        try:
            state, light_travel_time = spiceypy.spkezr(
                f"{self.spacecraft_code}",
                et,
                "J2000",
                "NONE",
                observer,
            )
        except spiceypy.SpiceSPKINSUFFDATA:
            raise BadEphemeris(
                "The time you have requested is outside of the time range where data exists for this spacecraft."
            )
        return np.asarray(state), np.asarray(light_travel_time)

    def get_spacecraft_position(
        self, time: Time, observer="SOLAR SYSTEM BARYCENTER"
    ) -> npt.NDArray:
        """Returns the position vector (x, y, z) in [km] for all `time` w.r.t the observer.

        Parameters:
        -----------
        time: astropy.time.Time
            Time array at which to estimate position. Time must be in UTC.
        observer: string
            Observer body. Common options include "SOLAR SYSTEM BARYCENTER", "EARTH BARYCENTER", "MOON BARYCENTER"
        """
        return self._get_state_vector(time=time, observer=observer)[0][:, :3]

    def get_spacecraft_velocity(
        self, time: Time, observer="SOLAR SYSTEM BARYCENTER"
    ) -> npt.NDArray:
        """Returns the position vector (vx, vy, vz) in [km/s] for all `time` w.r.t the observer.

        Parameters:
        -----------
        time: astropy.time.Time
            Time array at which to estimate velocity. Time must be in UTC.
        observer: string
            Observer body. Common options include "SOLAR SYSTEM BARYCENTER", "EARTH BARYCENTER", "MOON BARYCENTER"
        """
        return self._get_state_vector(time=time, observer=observer)[0][:, 3:]

    def get_spacecraft_light_travel_time(
        self, time: Time, observer="SOLAR SYSTEM BARYCENTER"
    ) -> npt.NDArray:
        """Returns the one-way light travel time in seconds for all `time` w.r.t the observer.

        Parameters:
        -----------
        time: astropy.time.Time
            Time array at which to estimate position. Time must be in UTC.
        observer: string
            Observer body. Common options include "SOLAR SYSTEM BARYCENTER", "EARTH BARYCENTER", "MOON BARYCENTER"
        """
        return self._get_state_vector(time=time, observer=observer)[1] / 2

    def get_barycentric_time_correction(
        self, time: Time, ra: Union[float, npt.NDArray], dec: Union[float, npt.NDArray]
    ) -> npt.NDArray:
        """Returns the barycentric time correction in days for observations of a particular target specified by RA and Dec.

        Note that `time` here must be time in spacecraft time, in UTC format.
        This means that for SPOC data this should be the time without the SPOC barycentric correction applied.

        Parameters:
        -----------
        time: astropy.time.Time
            Time array at which to estimate position. Time must be in UTC.
        ra: float, np.ndarray
            The right ascention of the target in degrees
        dec: float, np.ndarray
            The declination of the target in degrees
        """

        star_vector = np.array(
            [
                np.cos(np.deg2rad(dec)) * np.cos(np.deg2rad(ra)),
                np.cos(np.deg2rad(dec)) * np.sin(np.deg2rad(ra)),
                np.sin(np.deg2rad(dec)),
            ]
        )
        star_vector /= np.linalg.norm(star_vector, axis=0)
        position = self.get_spacecraft_position(time=time)
        return ((position * u.km).dot(star_vector) / (c)).to(u.s).value

    def get_velocity_aberrated_positions(
        self, time: Time, ra: float, dec: float
    ) -> npt.NDArray:
        """Returns the RA and Dec after velocity aberration has been applied.

        Note that `time` here must be time in spacecraft time, in UTC format.

        Parameters:
        -----------
        time: astropy.time.Time
            Time array at which to estimate position. Time must be in UTC.
        ra: float, np.ndarray
            The right ascention of the target in degrees
        dec: float, np.ndarray
            The declination of the target in degrees
        """

        # Compute the star vector (normalized direction vector for the target)
        star_vector = np.array(
            [
                np.cos(np.deg2rad(dec)) * np.cos(np.deg2rad(ra)),
                np.cos(np.deg2rad(dec)) * np.sin(np.deg2rad(ra)),
                np.sin(np.deg2rad(dec)),
            ]
        )

        # Normalize star_vector for safety (though it should already be normalized)
        star_vector /= np.linalg.norm(star_vector, axis=0)

        # Get the spacecraft velocity in m/s
        velocity = self.get_spacecraft_velocity(time=time) * 1000  # Convert km/s to m/s

        # Compute beta vector (velocity / speed of light)
        beta = velocity / c.value

        # Compute the scalar product beta \cdot star_vector
        beta_dot_star = np.sum(beta * star_vector, axis=-1)

        # Compute the relativistic factor gamma
        gamma = 1 / np.sqrt(1 - np.sum(beta**2, axis=-1))

        # Apply the aberration formula
        factor = 1 / (1 + beta_dot_star)

        star_vector_ab = factor[:, None] * (gamma[:, None] * star_vector + beta)

        # Normalize the aberrated vector
        star_vector_ab /= np.linalg.norm(star_vector_ab, axis=-1, keepdims=True)

        # Convert back to RA and Dec
        ra_aberrated = np.rad2deg(
            np.arctan2(star_vector_ab[:, 1], star_vector_ab[:, 0])
        )
        dec_aberrated = np.rad2deg(np.arcsin(star_vector_ab[:, 2]))

        # Ensure RA is in [0, 360] range
        ra_aberrated = np.mod(ra_aberrated, 360)

        # Reshape output to match input dimensions
        return (
            ra_aberrated,
            dec_aberrated,
        )


class KeplerSpacecraft(Spacecraft):
    """
    A class representing the Kepler spacecraft.

    This class extends the `Spacecraft` base class and includes spacecraft-specific
    configurations and calculations, such as correcting for timing errors unique to the
    Kepler mission.

    Attributes
    ----------
    spacecraft_code : int
        The SPICE NAIF ID code for the Kepler spacecraft.

    Methods
    -------
    get_barycentric_time_correction(time, ra, dec)
        Returns the barycentric time correction in days for observations of a target,
        applying mission-specific timing corrections.
    """

    spacecraft_code = -227
    time_offset = 2454833

    def get_barycentric_time_correction(
        self, time: Time, ra: Union[float, npt.NDArray], dec: Union[float, npt.NDArray]
    ) -> npt.NDArray:
        """Returns the barycentric time correction in days for observations of a particular target specified by RA and Dec.

        Note that `time` here must be time in spacecraft time, in UTC format.
        This means that for SPOC data this should be the time without the SPOC barycentric correction applied.

        Note this also corrects the timing error in the Kepler TIME column, see https://archive.stsci.edu/kepler/timing_error.html

        Parameters:
        -----------
        time: astropy.time.Time
            Time array at which to estimate position. Time must be in UTC.
        ra: float, np.ndarray
            The right ascention of the target in degrees
        dec: float, np.ndarray
            The declination of the target in degrees
        """
        time = self._process_time(time)
        tcorr = super().get_barycentric_time_correction(time, ra, dec)
        tcorr += 66.184
        k = time.jd > Time("2012-06-30 23:59:60", format="iso").jd
        tcorr[k] += 1
        return tcorr


class TESSSpacecraft(Spacecraft):
    """
    A class representing the TESS spacecraft.

    This class extends the `Spacecraft` base class and includes spacecraft-specific
    configurations, such as the unique SPICE NAIF ID code for the TESS mission.

    Attributes
    ----------
    spacecraft_code : int
        The SPICE NAIF ID code for the TESS spacecraft.
    """

    spacecraft_code = -95
    time_offset = 2457000
